@@ -39,6 +39,7 @@
 
 #include <libgda/sql-delimiter/gda-sql-delimiter.h>
 
+#define FILE_EXTENSION ".mdb"
 #define PARENT_TYPE GDA_TYPE_SERVER_PROVIDER
 
 #define OBJECT_DATA_MDB_HANDLE "GDA_Mdb_MdbHandle"
@@ -212,7 +213,10 @@ gda_mdb_provider_open_connection (GdaServerProvider *provider,
 				  const gchar *username,
 				  const gchar *password)
 {
-	const gchar *filename;
+	gchar *filename = NULL, *tmp;
+	const gchar *dirname = NULL, *dbname = NULL;
+	gchar *dup = NULL;
+
 	GdaMdbConnection *mdb_cnc;
 	GdaMdbProvider *mdb_prv = (GdaMdbProvider *) provider;
 
@@ -220,18 +224,74 @@ gda_mdb_provider_open_connection (GdaServerProvider *provider,
 	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), FALSE);
 
 	/* look for parameters */
-	filename = gda_quark_list_find (params, "FILENAME");
-	if (!filename) {
-		gda_connection_add_event_string (
-			cnc,
-			_("A database FILENAME must be specified in the connection_string"));
-		return FALSE;
-	}
+	dirname = gda_quark_list_find (params, "DB_DIR");
+	dbname = gda_quark_list_find (params, "DB_NAME");
+	
+	if (!dirname || !dbname) {
+                const gchar *str;
+
+                str = gda_quark_list_find (params, "FILENAME");
+                if (!str) {
+                        gda_connection_add_event_string (cnc,
+                                                         _("The connection string must contain DB_DIR and DB_NAME values"));
+                        return FALSE;
+                }
+                else {
+                        gint len = strlen (str);
+			gint elen = strlen (FILE_EXTENSION);
+
+			if (g_str_has_suffix (str, FILE_EXTENSION)) {
+                                gchar *ptr;
+
+                                dup = strdup (str);
+                                dup [len-elen] = 0;
+                                for (ptr = dup + (len - elen - 1); (ptr >= dup) && (*ptr != G_DIR_SEPARATOR); ptr--);
+                                dbname = ptr;
+                                if (*ptr == G_DIR_SEPARATOR)
+                                        dbname ++;
+
+                                if ((*ptr == G_DIR_SEPARATOR) && (ptr > dup)) {
+                                        dirname = dup;
+                                        while ((ptr >= dup) && (*ptr != G_DIR_SEPARATOR))
+                                                ptr--;
+                                        *ptr = 0;
+                                }
+                        }
+                        if (!dbname || !dirname) {
+                                gda_connection_add_event_string (cnc,
+                                                                 _("The connection string format has changed: replace FILENAME with "
+                                                                   "DB_DIR (the path to the database file) and DB_NAME "
+                                                                   "(the database file without the '%s' at the end)."), FILE_EXTENSION);
+				g_free (dup);
+                                return FALSE;
+                        }
+                        else
+                                g_warning (_("The connection string format has changed: replace FILENAME with "
+                                             "DB_DIR (the path to the database file) and DB_NAME "
+                                             "(the database file without the '%s' at the end)."), FILE_EXTENSION);
+                }
+        }
+
+	if (!g_file_test (dirname, G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR)) {
+                gda_connection_add_event_string (cnc,
+                                                 _("The DB_DIR part of the connection string must point "
+                                                   "to a valid directory"));
+                return FALSE;
+        }
+
+	tmp = g_strdup_printf ("%s%s", dbname, FILE_EXTENSION);
+        filename = g_build_filename (dirname, tmp, NULL);
+	g_free (dup);
+	g_free (tmp);
 
 	mdb_cnc = g_new0 (GdaMdbConnection, 1);
 	mdb_cnc->cnc = cnc;
 	mdb_cnc->server_version = NULL;
+#ifdef MDB_WITH_WRITE_SUPPORT
 	mdb_cnc->mdb = mdb_open (filename, MDB_WRITABLE);
+#else
+	mdb_cnc->mdb = mdb_open (filename);
+#endif
 	if (!mdb_cnc->mdb) {
 		gda_connection_add_event_string (cnc, _("Could not open file %s"), filename);
 		g_free (mdb_cnc);
@@ -793,7 +853,13 @@ gda_mdb_provider_execute_sql (GdaMdbProvider *mdbprv, GdaConnection *cnc, const
 		/* allocate bound data */
 		bound_data[c] = (gchar *) malloc (MDB_BIND_SIZE);
 		bound_data[c][0] = '\0';
+
+#ifdef MDB_SQL_BIND_COLUMN_FOUR_ARGS
+		int len = 0; /* Maybe we should actually do something with this output parameter. */
+		mdb_sql_bind_column (mdb_SQL, c + 1, bound_data[c], &len);
+#else
 		mdb_sql_bind_column (mdb_SQL, c + 1, bound_data[c]);
+#endif
 		
 		/* set description for the field */
 		fa = gda_data_model_describe_column (model, c);
@@ -871,7 +937,7 @@ gda_mdb_provider_get_data_handler (GdaServerProvider *provider,
 		 (type == GDA_TYPE_BLOB)) {
 		dh = gda_server_provider_handler_find (provider, cnc, type, NULL);
 		if (!dh) {
-			dh = gda_handler_bin_new_with_prov (provider, cnc);
+			dh = gda_handler_bin_new ();
 			if (dh) {
 				gda_server_provider_handler_declare (provider, dh, cnc, GDA_TYPE_BINARY, NULL);
 				gda_server_provider_handler_declare (provider, dh, cnc, GDA_TYPE_BLOB, NULL);
