@@ -1,11 +1,12 @@
 /* GDA Postgres provider
- * Copyright (C) 1998-2002 The GNOME Foundation
+ * Copyright (C) 1998 - 2005 The GNOME Foundation
  *
  * AUTHORS:
  *      Michael Lausch <michael@lausch.at>
  *	Rodrigo Moya <rodrigo@gnome-db.org>
  *      Vivien Malerba <malerba@gnome-db.org>
  *      Gonzalo Paniagua Javier <gonzalo@gnome-db.org>
+ *      Bas Driessen <bas.driessen@xobas.com>
  *
  * This Library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Library General Public License as
@@ -23,7 +24,9 @@
  * Boston, MA 02111-1307, USA.
  */
 
-#include <libgda/gda-intl.h>
+#include <glib/gi18n-lib.h>
+#include <libgda/gda-data-model.h>
+#include <libgda/gda-data-model-private.h>
 #include <string.h>
 #include "gda-postgres.h"
 #include "gda-postgres-recordset.h"
@@ -37,7 +40,7 @@
 struct _GdaPostgresRecordsetPrivate {
 	PGresult *pg_res;
 	GdaConnection *cnc;
-	GdaValueType *column_types;
+	GType *column_types;
 	gchar *table_name;
 	gint ncolumns;
 	gint nrows;
@@ -52,12 +55,13 @@ static void gda_postgres_recordset_init       (GdaPostgresRecordset *recset,
 					       GdaPostgresRecordsetClass *klass);
 static void gda_postgres_recordset_finalize   (GObject *object);
 
-static const GdaValue *gda_postgres_recordset_get_value_at    (GdaDataModel *model, gint col, gint row);
-static GdaFieldAttributes *gda_postgres_recordset_describe    (GdaDataModel *model, gint col);
-static gint gda_postgres_recordset_get_n_rows 		      (GdaDataModel *model);
-static const GdaRow *gda_postgres_recordset_get_row 	      (GdaDataModel *model, gint rownum);
-static const GdaRow *gda_postgres_recordset_append_row        (GdaDataModel *model, const GList *values);
-static gboolean gda_postgres_recordset_update_row 	      (GdaDataModel *model, const GdaRow *row);
+static const GValue *gda_postgres_recordset_get_value_at    (GdaDataModelRow *model, gint col, gint row);
+static void gda_postgres_recordset_describe_column    (GdaDataModelRow *model, gint col);
+static gint gda_postgres_recordset_get_n_rows 		      (GdaDataModelRow *model);
+static GdaRow *gda_postgres_recordset_get_row 	      (GdaDataModelRow *model, gint rownum, GError **error);
+static gboolean gda_postgres_recordset_append_row	      (GdaDataModelRow *model, GdaRow *row, GError **error);
+static gboolean gda_postgres_recordset_remove_row 	      (GdaDataModelRow *model, GdaRow *row, GError **error);
+static gboolean gda_postgres_recordset_update_row 	      (GdaDataModelRow *model, GdaRow *row, GError **error);
 
 static GObjectClass *parent_class = NULL;
 
@@ -81,16 +85,16 @@ static void
 gda_postgres_recordset_class_init (GdaPostgresRecordsetClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
-	GdaDataModelClass *model_class = GDA_DATA_MODEL_CLASS (klass);
+	GdaDataModelRowClass *model_class = GDA_DATA_MODEL_ROW_CLASS (klass);
 
 	parent_class = g_type_class_peek_parent (klass);
 
 	object_class->finalize = gda_postgres_recordset_finalize;
 	model_class->get_n_rows = gda_postgres_recordset_get_n_rows;
-	model_class->describe_column = gda_postgres_recordset_describe;
 	model_class->get_value_at = gda_postgres_recordset_get_value_at;
 	model_class->get_row = gda_postgres_recordset_get_row;
 	model_class->append_row = gda_postgres_recordset_append_row;
+	model_class->remove_row = gda_postgres_recordset_remove_row;
 	model_class->update_row = gda_postgres_recordset_update_row;
 }
 
@@ -108,19 +112,23 @@ gda_postgres_recordset_finalize (GObject * object)
 
 	g_free (recset->priv->column_types);
 	recset->priv->column_types = NULL;
+
+        g_free (recset->priv->table_name);
+        recset->priv->table_name = NULL;
+
 	g_free (recset->priv);
 	recset->priv = NULL;
 
 	parent_class->finalize (object);
 }
 
-static GdaValueType *
+static GType *
 get_column_types (GdaPostgresRecordsetPrivate *priv)
 {
 	gint i;
-	GdaValueType *types;
+	GType *types;
 
-	types = g_new (GdaValueType, priv->ncolumns);
+	types = g_new (GType, priv->ncolumns);
 	for (i = 0; i < priv->ncolumns; i++)
 		types [i] = gda_postgres_type_oid_to_gda (priv->type_data,
 							  priv->ntypes, 
@@ -130,12 +138,12 @@ get_column_types (GdaPostgresRecordsetPrivate *priv)
 }
 
 static GdaRow *
-get_row (GdaDataModel *model, GdaPostgresRecordsetPrivate *priv, gint rownum)
+get_row (GdaDataModel *model, GdaPostgresRecordsetPrivate *priv, gint rownum, GError **error)
 {
 	gchar *thevalue;
-	GdaValueType ftype;
+	GType ftype;
 	gboolean isNull;
-	GdaValue *value;
+	GValue *value;
 	GdaRow *row;
 	gint i;
 	gchar *id;
@@ -143,16 +151,12 @@ get_row (GdaDataModel *model, GdaPostgresRecordsetPrivate *priv, gint rownum)
 	
 	row = gda_row_new (model, priv->ncolumns);
 	for (i = 0; i < priv->ncolumns; i++) {
-		thevalue = PQgetvalue(priv->pg_res, rownum, i);
+		thevalue = PQgetvalue (priv->pg_res, rownum, i);
 		length = PQgetlength (priv->pg_res, rownum, i);
 		ftype = priv->column_types [i];
 		isNull = *thevalue != '\0' ? FALSE : PQgetisnull (priv->pg_res, rownum, i);
-		value = (GdaValue *) gda_row_get_value (row, i);
-		gda_postgres_set_value (value, ftype, thevalue, isNull, length);
-		if (gda_value_isa (value, GDA_VALUE_TYPE_BLOB)) {
-			GdaBlob *blob = (GdaBlob *) gda_value_get_blob (value);
-			gda_postgres_blob_set_connection (blob, priv->cnc);
-		}
+		value = (GValue *) gda_row_get_value (row, i);
+		gda_postgres_set_value (priv->cnc, value, ftype, thevalue, isNull, length);
 	}
 
 	gda_row_set_number (row, rownum);
@@ -167,8 +171,8 @@ get_row (GdaDataModel *model, GdaPostgresRecordsetPrivate *priv, gint rownum)
  * Overrides
  */
 
-static const GdaRow *
-gda_postgres_recordset_get_row (GdaDataModel *model, gint row)
+static GdaRow *
+gda_postgres_recordset_get_row (GdaDataModelRow *model, gint row, GError **error)
 {
 	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
 	GdaPostgresRecordsetPrivate *priv_data;
@@ -177,13 +181,14 @@ gda_postgres_recordset_get_row (GdaDataModel *model, gint row)
 	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), NULL);
 	g_return_val_if_fail (recset->priv != NULL, 0);
 
-	row_list = (GdaRow *) gda_data_model_hash_get_row (model, row);
+	row_list = (GdaRow *) GDA_DATA_MODEL_ROW_CLASS (parent_class)->get_row (model, row, 
+										error);
 	if (row_list != NULL)
-		return (const GdaRow *)row_list;
+		return row_list;
 
 	priv_data = recset->priv;
 	if (!priv_data->pg_res) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						 _("Invalid PostgreSQL handle"));
 		return NULL;
 	}
@@ -192,36 +197,35 @@ gda_postgres_recordset_get_row (GdaDataModel *model, gint row)
 		return NULL; /* For the last row don't add an error. */
 
 	if (row < 0 || row > priv_data->nrows) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Row number out of range"));
 		return NULL;
 	}
 
-	row_list = get_row (model, priv_data, row);
+	row_list = get_row (GDA_DATA_MODEL (model), priv_data, row, error);
 	gda_data_model_hash_insert_row (GDA_DATA_MODEL_HASH (model),
 					row, row_list);
+	g_object_unref (row_list); /* gda_data_model_hash_insert_row() holds a reference. */
 
-	return (const GdaRow *) row_list;
+	return row_list;
 }
 
-static const GdaRow *
-gda_postgres_recordset_append_row (GdaDataModel *model, const GList *values)
+static gboolean
+gda_postgres_recordset_append_row (GdaDataModelRow *model, GdaRow *row, GError **error)
 {
 	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
 	GdaPostgresRecordsetPrivate *priv_data;
-	GString *sql;
-	gint cols, i;
-	GList *l;
+	GString *sql, *sql_value;
+	gchar *cur_val;
+	gint i;
 	PGresult *pg_res;
 	GdaPostgresConnectionData *cnc_priv_data;
 	PGconn *pg_conn;
-	const GdaRow *row;
 
-	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), NULL);
-	g_return_val_if_fail (values != NULL, NULL);
-	g_return_val_if_fail (gda_data_model_is_updatable (model), NULL);
-	//g_return_val_if_fail (gda_data_model_hash_changed (model), NULL);
-	g_return_val_if_fail (recset->priv != NULL, NULL);
+	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), FALSE);
+	g_return_val_if_fail (row != NULL, FALSE);
+	g_return_val_if_fail (gda_data_model_is_updatable (GDA_DATA_MODEL (model)), FALSE);
+	g_return_val_if_fail (recset->priv != NULL, FALSE);
 	priv_data = recset->priv;
 	pg_res = priv_data->pg_res;
 	cnc_priv_data = g_object_get_data (G_OBJECT (priv_data->cnc),
@@ -230,98 +234,182 @@ gda_postgres_recordset_append_row (GdaDataModel *model, const GList *values)
 
 	/* checks if the table name has been guessed */
 	if (priv_data->table_name == NULL) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						 _("Table name could not be guessed."));
-		return NULL;
+		return FALSE;
 	}
 
-	cols = gda_data_model_get_n_columns (model);
-	if (cols != g_list_length ((GList *) values)) {
-		gda_connection_add_error_string (
-			priv_data->cnc,
-			_("Attempt to insert a row with an invalid number of columns"));
-		return NULL;
-	}
+	/* checks for correct number of columns */
+        if (priv_data->ncolumns != gda_row_get_length (row)) {
+                gda_connection_add_event_string (priv_data->cnc,
+                                                _("Attempt to insert a row with an invalid number of columns"));
+                return FALSE;
+        }
 
 	/* Prepare the SQL command */
 	sql = g_string_new ("INSERT INTO ");
-	g_string_append_printf (sql, "%s(",
-				priv_data->table_name);
-	for (i = 0; i < cols; i++) {
-		GdaFieldAttributes *fa;
+	g_string_append_printf (sql, "%s (",  priv_data->table_name);
 
-		fa = gda_data_model_describe_column (model, i);
-		if (!fa) {
-			gda_connection_add_error_string (
-				priv_data->cnc,
-				_("Could not retrieve column's information"));
-			g_string_free (sql, TRUE);
-			return NULL;
-		}
+	sql_value = g_string_new ("VALUES (");
+
+	for (i = 0; i < priv_data->ncolumns; i++) {
 
 		if (i != 0)
+		{
 			sql = g_string_append (sql, ", ");
-		sql = g_string_append (sql, gda_field_attributes_get_name (fa));
-	}
-	sql = g_string_append (sql, ") VALUES (");
-
-	for (l = (GList *) values, i = 0; i < cols; i++, l = g_list_next (l)) {
-		gchar *val_str;
-		const GdaValue *val = (const GdaValue *) l->data;
-	
-		if (!val) {
-			gda_connection_add_error_string (
-				priv_data->cnc,
-				_("Could not retrieve column's value"));
-			g_string_free (sql, TRUE);
-			return NULL;
+			sql_value = g_string_append (sql_value, ", ");
 		}
 
-		if (i != 0)
-			sql = g_string_append (sql, ", ");
-		val_str = gda_postgres_value_to_sql_string ((GdaValue *)val);
-		if (val_str == NULL) {
-			gda_connection_add_error_string (
-				priv_data->cnc,
-				_("Could not retrieve column's value"));
-			g_string_free (sql, TRUE);
-			return NULL;
-		}
-		sql = g_string_append (sql, val_str);
-		g_free (val_str);
+		sql = g_string_append (sql, "\"");
+		sql = g_string_append (sql, PQfname (priv_data->pg_res, i));
+		sql = g_string_append (sql, "\"");
+
+		cur_val = gda_value_stringify (gda_row_get_value ((GdaRow *) row, i));
+		sql_value = g_string_append (sql_value, "'");
+		sql_value = g_string_append (sql_value, cur_val);
+		sql_value = g_string_append (sql_value, "'");
+		g_free (cur_val);
 	}
+
+	/* concatenate SQL statement */
+        sql = g_string_append (sql, ") ");
+	sql = g_string_append (sql, sql_value->str);
 	sql = g_string_append (sql, ")");
 
 	/* execute the SQL query */
 	pg_res = PQexec (pg_conn, sql->str);
+
 	g_string_free (sql, TRUE);
+        g_string_free (sql_value, TRUE);
 
 	if (pg_res != NULL) {
 		/* update ok! */
 		if (PQresultStatus (pg_res) != PGRES_COMMAND_OK) {
-			gda_connection_add_error (priv_data->cnc,
-						  gda_postgres_make_error (pg_conn, pg_res));
+			gda_postgres_make_error (priv_data->cnc, pg_conn, pg_res);
 			PQclear (pg_res);
-			return NULL;
+			return FALSE;
 		}
 		PQclear (pg_res);
 	}
 	else
-		gda_connection_add_error (priv_data->cnc,
-					  gda_postgres_make_error (pg_conn, NULL));
-	row = GDA_DATA_MODEL_CLASS (parent_class)->append_row (model, values);
+		gda_postgres_make_error (priv_data->cnc, pg_conn, NULL);
 
-	return row;
+	/* append row in hash table */
+	if (! GDA_DATA_MODEL_ROW_CLASS (parent_class)->append_row (model, row, error)) {
+		gda_postgres_make_error (priv_data->cnc, pg_conn, pg_res);
+		return FALSE;
+	}
+
+	return TRUE;
 }
 
 static gboolean
-gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
+gda_postgres_recordset_remove_row (GdaDataModelRow *model, GdaRow *row, GError **error)
 {
 	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
 	GdaPostgresRecordsetPrivate *priv_data;
-	gint colnum, uk, nuk;
+	gint colnum, uk;
+	PGresult *pg_res, *pg_rm_res;
+	gchar *query, *query_where, *tmp;
+	GdaPostgresConnectionData *cnc_priv_data;
+	PGconn *pg_conn;
+	gboolean status = FALSE;
+
+	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), FALSE);
+	g_return_val_if_fail (recset->priv != NULL, FALSE);
+	g_return_val_if_fail (row != NULL, FALSE);
+
+	priv_data = recset->priv;
+	pg_res = priv_data->pg_res;
+	cnc_priv_data = g_object_get_data (G_OBJECT (priv_data->cnc),
+					   OBJECT_DATA_POSTGRES_HANDLE);
+	pg_conn = cnc_priv_data->pconn;
+
+	/* checks if the given row belongs to the given model */
+	if (gda_row_get_model ((GdaRow *) row) != GDA_DATA_MODEL (model)) {
+		gda_connection_add_event_string (priv_data->cnc,
+						_("Given row doesn't belong to the model."));
+		return FALSE;
+	}
+
+	/* checks if the table name has been guessed */
+	if (priv_data->table_name == NULL) {
+		gda_connection_add_event_string (priv_data->cnc,
+						_("Table name could not be guessed."));
+		return FALSE;
+	}
+
+	query_where = g_strdup ("WHERE TRUE ");
+
+	for (colnum = uk = 0;
+	     colnum != gda_data_model_get_n_columns (GDA_DATA_MODEL(model));
+	     colnum++)
+	{
+		GdaColumn *attrs = gda_data_model_describe_column (GDA_DATA_MODEL(model), colnum);
+		const gchar *column_name = PQfname (pg_res, colnum);
+		gchar *curval = gda_value_stringify (gda_row_get_value ((GdaRow *) row, colnum));
+
+		/* unique column: we will use it as an index */
+		if (gda_column_get_primary_key (attrs) ||
+		    gda_column_get_unique_key (attrs))
+		{
+			/* fills the 'where' part of the update command */
+			tmp = g_strdup_printf ("AND \"%s\" = '%s' ",
+					       column_name,
+					       curval);
+			query_where = g_strconcat (query_where, tmp, NULL);
+			g_free (tmp);
+			uk++;
+		}
+
+		g_free (curval);
+	}
+
+	if (uk == 0) {
+		gda_connection_add_event_string (priv_data->cnc,
+						_("Model doesn't have at least one unique key."));
+	}
+	else {
+		/* build the delete command */
+		query = g_strdup_printf ("DELETE FROM %s %s",
+					 priv_data->table_name,
+					 query_where);
+
+		/* remove the row */
+		pg_rm_res = PQexec (pg_conn, query);
+		g_free (query);
+
+		if (pg_rm_res != NULL) {
+			/* removal ok! */
+			if (PQresultStatus (pg_rm_res) == PGRES_COMMAND_OK)
+				status = TRUE;
+			else
+				gda_postgres_make_error (priv_data->cnc, pg_conn, pg_rm_res);
+			PQclear (pg_rm_res);
+		}
+		else
+			gda_postgres_make_error (priv_data->cnc, pg_conn, NULL);
+	}
+
+	g_free (query_where);
+
+	/* remove entry from data model */
+	if (status == TRUE)
+		status = GDA_DATA_MODEL_ROW_CLASS (parent_class)->remove_row (model, row, error);
+
+	return status;
+}
+
+static gboolean
+gda_postgres_recordset_update_row (GdaDataModelRow *model, GdaRow *row, GError **error)
+{
+	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
+	GdaPostgresRecordsetPrivate *priv_data;
+	gint colnum, rownum, uk, nuk;
 	PGresult *pg_res, *pg_upd_res;
 	gchar *query, *query_where, *query_set, *tmp;
+	gchar *oldval, *newval;
+	const gchar *column_name;
 	GdaPostgresConnectionData *cnc_priv_data;
 	PGconn *pg_conn;
 	gboolean status = FALSE;
@@ -336,35 +424,43 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 	pg_conn = cnc_priv_data->pconn;
 	
 	/* checks if the given row belongs to the given model */
-	if (gda_row_get_model ((GdaRow *) row) != model) {
-		gda_connection_add_error_string (priv_data->cnc,
+	if (gda_row_get_model ((GdaRow *) row) != GDA_DATA_MODEL (model)) {
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Given row doesn't belong to the model."));
 		return FALSE;
 	}
 
 	/* checks if the table name has been guessed */
 	if (priv_data->table_name == NULL) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Table name could not be guessed."));
 		return FALSE;
 	}
 
+	rownum = gda_row_get_number ((GdaRow *) row);
+
 	query_where = g_strdup ("WHERE TRUE ");
 	query_set = g_strdup ("SET ");
 
 	for (colnum = uk = nuk = 0;
-	     colnum != gda_data_model_get_n_columns (model);
+	     colnum != gda_data_model_get_n_columns (GDA_DATA_MODEL (model));
 	     colnum++) 
 	{
-		GdaFieldAttributes *attrs = gda_data_model_describe_column (model, colnum);
-		const gchar *column_name = PQfname (pg_res, colnum);
-		gchar *newval = gda_value_stringify (gda_row_get_value ((GdaRow *) row, colnum));
-		const gchar *oldval = PQgetvalue (pg_res, gda_row_get_number ((GdaRow *) row), colnum);
+		GdaColumn *attrs = gda_data_model_describe_column (GDA_DATA_MODEL (model), colnum);
+		column_name = PQfname (pg_res, colnum);
+		newval = gda_value_stringify (gda_row_get_value ((GdaRow *) row, colnum));
+
+		/* for data from mysql result we can retrieve original values to avoid
+		   unique columns to be updated */
+		if (rownum < priv_data->nrows)
+			oldval = PQgetvalue (pg_res, gda_row_get_number ((GdaRow *) row), colnum);
+		else
+			oldval = newval;
 
 		/* unique column: we won't update it, but we will use it as
 		   an index */
-		if (gda_field_attributes_get_primary_key (attrs) ||
-		    gda_field_attributes_get_unique_key (attrs)) 
+		if (gda_column_get_primary_key (attrs) ||
+		    gda_column_get_unique_key (attrs)) 
 		{
 			/* checks if it hasn't be modified anyway */
 			if (oldval == NULL ||
@@ -373,7 +469,7 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 			    	continue;
 
 			/* fills the 'where' part of the update command */
-			tmp = g_strdup_printf ("AND %s = '%s' ",
+			tmp = g_strdup_printf ("AND \"%s\" = '%s' ",
 					       column_name,
 					       newval);
 			query_where = g_strconcat (query_where, tmp, NULL);
@@ -383,7 +479,7 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 		/* non-unique column: update it */
 		else {
 			/* fills the 'set' part of the update command */
-			tmp = g_strdup_printf ("%s = '%s', ", 
+			tmp = g_strdup_printf ("\"%s\" = '%s', ", 
 					       column_name,
 					       newval);
 			query_set = g_strconcat (query_set, tmp, NULL);
@@ -392,15 +488,14 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 		}
 
 		g_free (newval);
-		gda_field_attributes_free (attrs);
 	}
 
 	if (uk == 0) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Model doesn't have at least one non-modified unique key."));
 	}
 	else if (nuk == 0) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Model doesn't have any non-unique values to update."));
 	}
 	else {
@@ -424,44 +519,42 @@ gda_postgres_recordset_update_row (GdaDataModel *model, const GdaRow *row)
 			if (PQresultStatus (pg_upd_res) == PGRES_COMMAND_OK)
 				status = TRUE;
 			else
-				gda_connection_add_error (priv_data->cnc,
-			        	                  gda_postgres_make_error (pg_conn, pg_upd_res));
+				gda_postgres_make_error (priv_data->cnc, pg_conn, pg_upd_res);
 			PQclear (pg_upd_res);
 		}
 		else
-			gda_connection_add_error (priv_data->cnc,
-			       	                  gda_postgres_make_error (pg_conn, NULL));
+			gda_postgres_make_error (priv_data->cnc, pg_conn, NULL);
 	}
 	
 	g_free (query_set);
 	g_free (query_where);
 
-	if (status == TRUE)
-               status = GDA_DATA_MODEL_CLASS (parent_class)->update_row (model, row);
+	/* emit update signals */
+	gda_data_model_row_updated (GDA_DATA_MODEL (model), gda_row_get_number ((GdaRow *) row));
 
 	return status;
 }
 
-static const GdaValue *
-gda_postgres_recordset_get_value_at (GdaDataModel *model, gint col, gint row)
+static const GValue *
+gda_postgres_recordset_get_value_at (GdaDataModelRow *model, gint col, gint row)
 {
 	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
 	GdaPostgresRecordsetPrivate *priv_data;
 	PGresult *pg_res;
 	GdaRow *row_list;
-	const GdaValue *value;
+	const GValue *value;
 
 	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), NULL);
 	g_return_val_if_fail (recset->priv != NULL, 0);
 	
-	value = gda_data_model_hash_get_value_at (model, col, row);
+	value = GDA_DATA_MODEL_ROW_CLASS (parent_class)->get_value_at (model, col, row);
 	if (value != NULL)
 		return value;
 
 	priv_data = recset->priv;
 	pg_res = priv_data->pg_res;
 	if (!pg_res) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						 _("Invalid PostgreSQL handle"));
 		return NULL;
 	}
@@ -470,20 +563,22 @@ gda_postgres_recordset_get_value_at (GdaDataModel *model, gint col, gint row)
 		return NULL; /* For the last row don't add an error. */
 
 	if (row < 0 || row > priv_data->nrows) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Row number out of range"));
 		return NULL;
 	}
 
 	if (col >= priv_data->ncolumns) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Column number out of range"));
 		return NULL;
 	}
-
-	row_list = get_row (model, priv_data, row);
+	
+	row_list = get_row (GDA_DATA_MODEL (model), priv_data, row, NULL);
 	gda_data_model_hash_insert_row (GDA_DATA_MODEL_HASH (model),
 					 row, row_list);
+	g_object_unref (row_list); /* gda_data_model_hash_insert_row() holds a reference. */
+
 	return gda_row_get_value (row_list, col);
 }
 
@@ -508,13 +603,17 @@ static gchar *guess_table_name (GdaPostgresRecordset *recset)
 	pg_conn = cnc_priv_data->pconn;
 
 	if (PQnfields (pg_res) > 0) {
-		gchar *query = g_strdup_printf ("SELECT c.relname FROM pg_catalog.pg_class c WHERE c.relkind = 'r'");
+		gchar *query = g_strdup_printf ("SELECT c.relname FROM pg_catalog.pg_class c WHERE c.relkind = 'r' AND c.relnatts = '%d'", PQnfields (pg_res));
 		gint i;
 		for (i = 0; i < PQnfields (pg_res); i++) {
 			const gchar *column_name = PQfname (pg_res, i);
 			gchar *cond = g_strdup_printf (" AND '%s' IN (SELECT a.attname FROM pg_catalog.pg_attribute a WHERE a.attrelid = c.oid)", column_name);
-			query = g_strconcat (query, cond, NULL);
+			gchar *tmp_query = NULL;
+
+			tmp_query = g_strconcat (query, cond, NULL);
+			g_free (query);
 			g_free (cond);
+			query = tmp_query;
 		}
 		pg_name_res = PQexec (pg_conn, query);
 		if (pg_name_res != NULL) {
@@ -536,7 +635,7 @@ static gchar *guess_table_name (GdaPostgresRecordset *recset)
  *    'u': unique key
  *    etc...
  */
-static gboolean check_constraint (const GdaDataModel *model,
+static gboolean check_constraint (const GdaDataModelRow *model,
 				  const gchar *table_name,
 				  const gint col,
 				  const gchar contype)
@@ -570,54 +669,54 @@ static gboolean check_constraint (const GdaDataModel *model,
 	return state;
 }
 
-static GdaFieldAttributes *
-gda_postgres_recordset_describe (GdaDataModel *model, gint col)
+static void
+gda_postgres_recordset_describe_column (GdaDataModelRow *model, gint col)
 {
 	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
 	GdaPostgresRecordsetPrivate *priv_data;
 	PGresult *pg_res;
-	GdaValueType ftype;
+	GType ftype;
 	gint scale;
-	GdaFieldAttributes *field_attrs;
+	GdaColumn *field_attrs;
 	gboolean ispk = FALSE, isuk = FALSE;
 
-	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (recset), NULL);
-	g_return_val_if_fail (recset->priv != NULL, 0);
+	g_return_if_fail (GDA_IS_POSTGRES_RECORDSET (recset));
+	g_return_if_fail (recset->priv != NULL);
 
 	priv_data = recset->priv;
 	pg_res = priv_data->pg_res;
 	if (!pg_res) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Invalid PostgreSQL handle"));
-		return NULL;
+		return;
 	}
 
 	if (col >= priv_data->ncolumns) {
-		gda_connection_add_error_string (priv_data->cnc,
+		gda_connection_add_event_string (priv_data->cnc,
 						_("Column out of range"));
-		return NULL;
+		return;
 	}
 
-	field_attrs = gda_field_attributes_new ();
-	gda_field_attributes_set_name (field_attrs, PQfname (pg_res, col));
+	field_attrs = gda_data_model_describe_column (GDA_DATA_MODEL (model), col);
+	gda_column_set_name (field_attrs, PQfname (pg_res, col));
 
 	ftype = gda_postgres_type_oid_to_gda (priv_data->type_data,
 					      priv_data->ntypes, 
 					      PQftype (pg_res, col));
 
-	scale = (ftype == GDA_VALUE_TYPE_DOUBLE) ? DBL_DIG :
-		(ftype == GDA_VALUE_TYPE_SINGLE) ? FLT_DIG : 0;
+	scale = (ftype == G_TYPE_DOUBLE) ? DBL_DIG :
+		(ftype == G_TYPE_FLOAT) ? FLT_DIG : 0;
 
-	gda_field_attributes_set_scale (field_attrs, scale);
-	gda_field_attributes_set_gdatype (field_attrs, ftype);
+	gda_column_set_scale (field_attrs, scale);
+	gda_column_set_g_type (field_attrs, ftype);
 
 	/* PQfsize() == -1 => variable length */
-	gda_field_attributes_set_defined_size (field_attrs,
+	gda_column_set_defined_size (field_attrs,
 					       PQfsize (pg_res, col));
 					       
-	gda_field_attributes_set_references (field_attrs, "");
+	gda_column_set_references (field_attrs, "");
 
-	gda_field_attributes_set_table (field_attrs,
+	gda_column_set_table (field_attrs,
 					priv_data->table_name);
 
 	if (priv_data->table_name) {
@@ -632,15 +731,13 @@ gda_postgres_recordset_describe (GdaDataModel *model, gint col)
 					 'u');
 	}
 	
-	gda_field_attributes_set_primary_key (field_attrs, ispk);
-	gda_field_attributes_set_unique_key (field_attrs, isuk);
+	gda_column_set_primary_key (field_attrs, ispk);
+	gda_column_set_unique_key (field_attrs, isuk);
 	/* FIXME: set_allow_null? */
-
-	return field_attrs;
 }
 
 static gint
-gda_postgres_recordset_get_n_rows (GdaDataModel *model)
+gda_postgres_recordset_get_n_rows (GdaDataModelRow *model)
 {
 	gint parent_row_num;
 	GdaPostgresRecordset *recset = (GdaPostgresRecordset *) model;
@@ -648,8 +745,13 @@ gda_postgres_recordset_get_n_rows (GdaDataModel *model)
 	g_return_val_if_fail (GDA_IS_POSTGRES_RECORDSET (model), 0);
 	g_return_val_if_fail (recset->priv != NULL, 0);
 
-	parent_row_num = GDA_DATA_MODEL_CLASS (parent_class)->get_n_rows (model);
-	return MAX(recset->priv->nrows, parent_row_num);
+	parent_row_num = GDA_DATA_MODEL_ROW_CLASS (parent_class)->get_n_rows (model);
+
+	/* if not initialized return number of PQ Tuples */
+	if (parent_row_num < 0)
+		return recset->priv->nrows;
+	else
+		return parent_row_num;
 }
 
 /*
@@ -685,6 +787,7 @@ gda_postgres_recordset_new (GdaConnection *cnc, PGresult *pg_res)
 	GdaPostgresConnectionData *cnc_priv_data;
 	gchar *cmd_tuples;
 	gchar *endptr [1];
+	gint i;
 
 	g_return_val_if_fail (GDA_IS_CONNECTION (cnc), NULL);
 	g_return_val_if_fail (pg_res != NULL, NULL);
@@ -699,18 +802,23 @@ gda_postgres_recordset_new (GdaConnection *cnc, PGresult *pg_res)
 	model->priv->h_table = cnc_priv_data->h_table;
 	model->priv->ncolumns = PQnfields (pg_res);
 	cmd_tuples = PQcmdTuples (pg_res);
-	if (cmd_tuples == NULL || *cmd_tuples == '\0'){
+	if (cmd_tuples == NULL || *cmd_tuples == '\0')
 		model->priv->nrows = PQntuples (pg_res);
-	} else {
+	else {
 		model->priv->nrows = strtol (cmd_tuples, endptr, 10);
 		if (**endptr != '\0')
 			g_warning (_("Tuples:\"%s\""), cmd_tuples);
-
+		
 	}
 	model->priv->column_types = get_column_types (model->priv);
 	gda_data_model_hash_set_n_columns (GDA_DATA_MODEL_HASH (model),
-					    model->priv->ncolumns);
+					   model->priv->ncolumns);
 	model->priv->table_name = guess_table_name (model);
+
+	/* set GdaColumn attributes */
+	for (i = 0; i < model->priv->ncolumns; i++)
+		gda_postgres_recordset_describe_column (GDA_DATA_MODEL_ROW (model), i);
+
 	return GDA_DATA_MODEL (model);
 }
 
